#include <iostream>
#include <cstdlib>
#include <string>
#include <cstring>
#include <sstream>
#include <fstream>
#include <map>
#include <vector>
#include "fasta.h"
#include "utils.h"
#include "tabix/tabix.hpp"

using namespace std;

const short BREAK = 1;
const short REPLACE_WITH_NS = 2;

const string ERROR_PREFIX = "[REAPR break] ";

string int2string(int number);

struct Breakpoint
{
    unsigned long start;
    unsigned long end;
    short type;
};

struct CmdLineOptions
{
    double minTriError;
    unsigned long minScaffLength;
    string fastaIn;
    string outprefix;
    string gff;
};

// deals with command line options: fills the options struct
void parseOptions(int argc, char** argv, CmdLineOptions& ops);

int main(int argc, char** argv)
{
    CmdLineOptions options;
    parseOptions(argc, argv, options);
    Fasta seq;
    string line;
    map<string, vector<Breakpoint> > breakpoints;
    ifstream inStream;
    ofstream outStream, outStreamBin;
    Tabix ti(options.gff);
    string fastaOut = options.outprefix + ".broken_assembly.fa";
    string binOut = options.outprefix + ".broken_assembly_bin.fa";
    string infoOut = options.outprefix + ".info.txt";

    outStream.open(infoOut.c_str());
    if (! outStream.is_open())
    {
        cerr << ERROR_PREFIX << "Error opening file '" << infoOut << "'" << endl;
        return 1;
    }

    // put breakpoints into memory.  coords in file are 1-based
    while (ti.getNextLine(line))
    {
        vector<string> v;
        Breakpoint bp;
        split(line, '\t', v);
        string id = v[0];

        // we break when errors are called over a gap, replace with Ns when
        // error called in a contig (i.e. not over a gap)
        if (v[2].compare("Frag_cov_gap") == 0)
        {
            // scaffold19_size430204    REAPR   Frag_cov_gap    362796  363397  0.00166113  .   .   Note=Error: Fragment coverage too low over gap 363104-363113;colour=12
            vector<string> a,b,c;
            split(v.back(), ';', a);
            split(a[a.size() - 2], ' ', b);
            split(b.back(), ',', c);
            for (unsigned long i = 0; i < c.size(); i++)
            {
                vector<string> d;
                split(c[i], '-', d);
                bp.start = atoi(d[0].c_str());
                bp.end = atoi(d[1].c_str());
                bp.type = BREAK;
                outStream << id << '\t' << bp.start << '\t' << bp.end << '\t' << v[2] << '\t' << "BREAK\n";;
                bp.start -= 2;
                breakpoints[id].push_back(bp);
            }
        }
        else if (v[2].compare("FCD_gap") == 0 && atof(v[5].c_str()) >= options.minTriError)
        {
            // scaffold7_size612284 REAPR   FCD_gap 547103  553357  0.781932    .   .   Note=Error: FCD failure over gap 550388-550397,552547-552686;colour=16
            vector<string> a,b,c;
            split(v.back(), ';', a);
            split(a[a.size() - 2], ' ', b);
            split(b.back(), ',', c);
            for (unsigned long i = 0; i < c.size(); i++)
            {
                vector<string> d;
                split(c[i], '-', d);
                bp.start = atoi(d[0].c_str());
                bp.end = atoi(d[1].c_str());
                bp.type = BREAK;
                outStream << id << '\t' << bp.start << '\t' << bp.end << '\t' << v[2] << '\t' << "BREAK\n";;
                bp.start -= 2;
                breakpoints[id].push_back(bp);
            }
        }
        else if (v[2].compare("Frag_cov") == 0 || (v[2].compare("FCD") == 0 && atof(v[5].c_str()) >= options.minTriError))
        {
            // scaffold6_size716595 REAPR   Frag_cov    600296  601454  0   .   .   Note=Error: Fragment coverage too low;color=15
            bp.start = atoi(v[3].c_str()) - 1;
            bp.end = atoi(v[4].c_str()) - 1;
            bp.type = REPLACE_WITH_NS;
            breakpoints[id].push_back(bp);
            outStream << id << '\t' << bp.start << '\t' << bp.end << '\t' << v[2] << "\tNs\n";;
        }
    }

    outStream.close();
    inStream.open(options.fastaIn.c_str());

    if (! inStream.is_open())
    {
        cerr << ERROR_PREFIX << "Error opening file '" << options.fastaIn << "'" << endl;
        return 1;
    }

    outStream.open(fastaOut.c_str());

    if (! outStream.is_open())
    {
        cerr << ERROR_PREFIX << "Error opening file '" << fastaOut << "'" << endl;
        return 1;
    }

    outStreamBin.open(binOut.c_str());

    if (! outStreamBin.is_open())
    {
        cerr << ERROR_PREFIX << "Error opening file '" << binOut << "'" << endl;
        return 1;
    }

    // do the breaking
    while (seq.fillFromFile(inStream))
    {
        map<string, vector<Breakpoint> >::iterator p;
        p = breakpoints.find(seq.id);
        vector<unsigned long> starts, ends;
        unsigned long counter = 1;

        if (p ==  breakpoints.end())
        {
            if (seq.length() >= options.minScaffLength)
                seq.print(outStream, 60);
            continue;
        }

        // get the start and end points
        for (vector<Breakpoint>::iterator q = p->second.begin(); q != p->second.end(); q++)
        {
            // 'replace with Ns' error
            if (q->type == REPLACE_WITH_NS)
            {
                stringstream ss;
                Fasta contig = seq.subseq(q->start, q->end);
                ss << q->start + 1 << '-' << q->end + 1;
                contig.id += ":" +  ss.str();
                contig.print(outStreamBin, 60);
                seq.seq.replace(q->start, q->end - q->start + 1, q->end - q->start + 1, 'N');
            }
            // delete a section and break
            else if (q->type == BREAK)
            {
                starts.push_back(q->start);
                ends.push_back(q->end);
            }
        }

        if (starts.size() == 0)
        {
            if (seq.length() >= options.minScaffLength)
                seq.print(outStream, 60);
            continue;
        }

        // now make the broken sequences
        if (starts[0] >= options.minScaffLength)
        {
            string originalID = seq.id;
            seq.id = seq.id + "." + int2string(counter) + ".1-" + int2string(starts[0] + 1);
            seq.subseq(0, starts[0]).print(outStream, 60);
            seq.id = originalID;
            counter++;
        }

        for (unsigned int i = 0; i < starts.size() - 1; i++)
        {
            if (starts[i+1] -ends[i] + 1 >= options.minScaffLength)
            {
                string originalID = seq.id;
                seq.id += "." + int2string(counter) + "." + int2string(ends[i] + 1) + "-" + int2string(starts[i+1] + 1);
                seq.subseq(ends[i], starts[i+1]).print(outStream, 60);
                seq.id = originalID;
                counter++;
            }
        }

        if (ends.back() != seq.length() - 1 && seq.length() - ends.back() >= options.minScaffLength)
        {
            string originalID = seq.id;
            seq.id += "." + int2string(counter) + "." + int2string(ends.back() + 1) + "-" +  int2string(seq.length());
            seq.subseq(ends.back(), seq.length() - 1).print(outStream, 60);
            seq.id = originalID;
        }
    }

    inStream.close();
    outStream.close();
    outStreamBin.close();

    return 0;
}

void parseOptions(int argc, char** argv, CmdLineOptions& ops)
{
    string usage;
    short requiredArgs = 3;
    int i;
    usage = "\
where 'errors.gff.gz' is the errors gff file made when running score.\n\n\
Options:\n\
-e <float>\n\tMinimum insert size error [0]\n\
-l <int>\n\tMinimum sequence length to output [1]\n\
";

    if (argc == 2 && strcmp(argv[1], "--wrapperhelp") == 0)
    {
        usage = "[options] <assembly.fa> <errors.gff.gz> <outfiles prefix>\n\n" + usage;
        cerr << usage << endl;
        exit(1);
    }
    else if (argc < requiredArgs)
    {
        usage = "[options] <errors.gff.gz> <assembly.fa> <outfiles prefix>\n\n" + usage;
        cerr << "usage: task_break " << usage;
        exit(1);
    }

    ops.minTriError = 0;
    ops.minScaffLength = 1;

    for (i = 1; i < argc - requiredArgs; i++)
    {
        if (strcmp(argv[i], "-e") == 0)
        {
            ops.minTriError = atof(argv[i+1]);
        }
        else if (strcmp(argv[i], "-l") == 0)
        {
            ops.minScaffLength = atoi(argv[i+1]);
        }
        else
        {
            cerr << ERROR_PREFIX << "Error! Switch not recognised: " << argv[i] << endl;
            exit(1);
        }
        i++;
    }

    if (argc - i != requiredArgs || argv[i+1][0] == '-')
    {
        cerr << usage;
        exit(1);
    }

    ops.fastaIn = argv[i];
    ops.gff = argv[i+1];
    ops.outprefix = argv[i+2];
}


string int2string(int number)
{
    stringstream ss;
    ss << number;
    return ss.str();
}
